from enum import IntFlag
from dataclasses import dataclass, field
import dataclasses
from typing import List
import logging
import json

class TrackingType(IntFlag):
    DEFAULT = 0
    WEART_HAND = 1

class HandSide(IntFlag):
	#HSnone = 0	
	Left = 1 << 0
	Right = 1 << 1
	

class HandClosingState(IntFlag):
	Open = 0
	Closing = 1
	Closed = 2

class GraspingState(IntFlag):
	Grapped = 0
	Released = 1

class ActuationPoint(IntFlag):
	#APnone	= 0
	Thumb = 1 << 0
	Index = 1 << 1
	Middle = 1 << 2
	Palm = 1 << 3


class CalibrationStatus(IntFlag):
	IDLE = 0
	Calibrating = 1
	Running = 2

class MiddlewareStatus(IntFlag):
	DISCONNECTED=0
	IDLE = 1
	STARTING = 2
	RUNNING = 3
	STOPPING = 4
	UPLOADING_TEXTURES = 5
	CONNECTING_DEVICE = 6
	CALIBRATION = 7
	def __str__(self):
		return self.name

class TextureType(IntFlag):
	ClickNormal = 0
	ClickSoft = 1
	DoubleClick = 2
	AluminiumFineMeshSlow = 3
	AluminiumFineMeshFast = 4
	PlasticMeshSlow = 5
	ProfiledAluminiumMeshMedium = 6
	ProfiledAluminiumMeshFast = 7
	RhombAluminiumMeshMedium = 8
	TextileMeshMedium = 9
	CrushedRock = 10
	VenetianGranite = 11
	SilverOak = 12
	LaminatedWood = 13
	ProfiledRubberSlow = 14
	VelcroHooks = 15
	VelcroLoops = 16
	PlasticFoil2 = 17
	Leather = 18
	Cotton1 = 19
	Aluminium = 20
	DoubleSidedTape = 21

@dataclass
class AccelerometerData:
	x: float = 0.0
	y: float = 0.0
	z: float = 0.0


@dataclass
class GyroscopeData:
	x: float = 0.0
	y: float = 0.0
	z: float = 0.0

@dataclass
class TofData:
	distance: int = 0

@dataclass
class SensorData:
	accelerometer: AccelerometerData
	gyroscope: GyroscopeData
	timeOfFlight: TofData

@dataclass
class AnalogSensorRawData:
	ntcTemperatureRaw:float
	ntcTemperatureConverted:float
	forceSensingRaw:float
	forceSensingConverted:float

@dataclass
class MiddlewareConnectedDevice:
	macAddress:str
	handSide:HandSide

@dataclass
class MiddlewareStatusData:
	status:MiddlewareStatus
	version:str
	statusCode:int
	errorDesc:str
	actuationsEnabled:bool
	connectedDevices:list #std::vector<MiddlewareConnectedDevice>
	timestamp:int = 0

@dataclass
class ThimbleStatus:
	id:ActuationPoint = ActuationPoint.Palm
	connected:bool = False
	statusCode:int = 0
	errorDesc:str = ""

@dataclass
class ConnectedDeviceStatus:
	macAddress:str = ""
	handSide: HandSide = HandSide.Left
	batteryLevel: int = 0
	charging: bool = False
	thimbles: List[ThimbleStatus] = field(default_factory=lambda: [ThimbleStatus()]) #std::vector<ThimbleStatus> thimbles;


def dataclass_from_list(klass, l):
	ret = []
	for elem in l:
		ret.append(dataclass_from_dict(klass, elem))
	return ret

def dataclass_from_dict(klass, d):
	if dataclasses.is_dataclass(klass):
		fieldtypes = {f.name:f.type for f in dataclasses.fields(klass)}
		k = klass(**{f:dataclass_from_dict(fieldtypes[f],d[f]) for f in d})
		return k
	else:
		if klass == MiddlewareStatus:
			return MiddlewareStatus[d]
		if klass == ActuationPoint:
			return ActuationPoint[str(d).capitalize()]
		if klass == HandSide:
			return HandSide[str(d).capitalize()]
		if klass == List[ThimbleStatus]:
			return dataclass_from_list(ThimbleStatus, d)
		return d # Not a dataclass field

def dict_from_dataclass(k):
	#to be checked, IntFlags are not well represented
	if dataclasses.is_dataclass(k):
		return  dataclasses.asdict(k)
	elif isinstance(k, list):
		l = []
		for elem in k:
			l.append(dict_from_dataclass(elem))
		return l



DEFAULT_IP_ADDRESS = "127.0.0.1"
DEFAULT_TCP_PORT = 13031

WEART_SDK_VERSION = "1.0.3"
WEART_SDK_TYPE = "SdkLLCPP"

defaultTemperature = 0.5
minTemperature = 0.0
maxTemperature = 1.0

defaultForce = 0.0
minForce = 0.0
maxForce = 1.0

defaultClosure = 0.0
minClosure = 0.0
maxClosure = 1.0

defaultAbduction = 0.442

defaultTextureIndex = 0
minTextureIndex = 0
maxTextureIndex = 21
nullTextureIndex = 255

defaultCollisionMultiplier = 20.0
minCollisionMultiplier = 0.0
maxCollisionMultiplier = 100.0

defaultVolumeTexture = 100.0
minVolumeTexture = 0.0
maxVolumeTexture = 100.0

thresholdThumbClosure = 0.5
thresholdIndexClosure = 0.5
thresholdMiddleClosure = 0.5

defaultGraspForce = 0.3
dynamicForceSensibility = 10.0

__all__ = [] # so we do not fill up the imports



import socket
import sys
import signal
import time
from threading import Thread
from enum import Enum
import logging

from . import WeArtCommon
from .WeArtCommon import TrackingType
from . import WeArtMessages as WeArtMessages
from .WeArtMessageSerializer import WeArtMessageSerializer
from  .WeArtThimbleTrackingObject import WeArtThimbleTrackingObject
from .WeArtMessageListener import WeArtMessageListener
from .WeArtTrackingRawData import WeArtTrackingRawData
from .WeArtAnalogSensorData import WeArtAnalogSensorData


logging.basicConfig()

class WeArtClient:
    messagesSeparator = '~'

    def __init__(self, ip_address, port, log_level = logging.DEBUG):
        self._messageSerializer = WeArtMessageSerializer()
        self.__Connected = False
        self.__Closing = False
        self.__s = None #socket
        self.__thimbleTrackingObjects = []
        self.__messageListeners = []
        self.__messageCallbacks = []
        self.__connectionStatusCallbacks = []
        self.__errorCallbacks = []
        self.__pendingCallbacks = []
        self.__IP_ADDRESS = ip_address
        self.__PORT = port
        self.__logger = logging.getLogger("WeArtClient")
        self.__logger.setLevel(log_level)

    class ErrorType(Enum):
        ConnectionError = 0
        SendMessageError = 1
        ReceiveMessageError = 2

    def Start(self, tracking_type = TrackingType.WEART_HAND):
        start_msg = None
        if tracking_type != None:
            start_msg = WeArtMessages.StartFromClientMessage(trackType=tracking_type)
        else:
            start_msg = WeArtMessages.StartFromClientMessage()
        self.SendMessage(start_msg)
        
    def Stop(self):
        stop_msg = WeArtMessages.StopFromClientMessage()
        self.SendMessage(stop_msg)
    
    def Run(self):
        try:
            self.__s = socket.socket()
            server_addr = (self.__IP_ADDRESS, self.__PORT)
            self.__s.connect(server_addr)
            self.__logger.info(f"Connection to server: { server_addr } established.")
            self.__Connected = True
            self.__NotifyConnectionStatus(True)
            t = Thread(target=self._OnReceive, args = [], daemon=True)
            t.start()

            #getStatusMessage = WeArtMessages.GetMiddlewareStatus()
            #self.SendMessage(getStatusMessage)

            getDevicesMessage = WeArtMessages.GetDevicesStatusMessage()
            self.SendMessage(getDevicesMessage)

        except socket.error as e:
            self.__logger.error(f"Unable to connect to server { server_addr }... \n{e}")
            self.__Connected = False
            self.__NotifyError(self.ErrorType.ConnectionError)
            sys.exit()
        return
    
    def IsConnected(self):
        return self.__Connected
    
    def Close(self):
        self.__Closing = True
        self.__s.close()
        self.__Connected = False
        self.__NotifyConnectionStatus(False)

    def StartCalibration(self):
        startCalibration = WeArtMessages.StartCalibrationMessage()
        self.SendMessage(startCalibration)

    def StopCalibration(self):
        stopCalibration = WeArtMessages.StopCalibrationMessage()
        self.SendMessage(stopCalibration)

    def StartRawData(self):
        message = WeArtMessages.RawDataOn()
        self.SendMessage(message)

    def StopRawData(self):
        message = WeArtMessages.RawDataOff()
        self.SendMessage(message)
    
    def SendMessage(self, msg:WeArtMessages.WeArtMessage):
        if not self.__Connected:
            return
        if msg == None:
            return
        
        text = msg.serialize()
        text+= self.messagesSeparator

        self.__logger.debug(f"Message to be sent: { text }")

        bytes = self.__s.send(text.encode())
        if bytes == 0:
            self.__Connected = False
            self.__logger.error(f"Send message '{ text }' failed")
            self.__s.close()
            self.__NotifyError(self.ErrorType.SendMessageError)
            self.__NotifyConnectionStatus(False)
            return


    def AddThimbleTracking(self, trackingObjects: WeArtThimbleTrackingObject):
        self.__thimbleTrackingObjects.append(trackingObjects)
        self.AddMessageListener(trackingObjects)

    def AddThimbleRawSensors(self, rawSensorData: WeArtTrackingRawData):
        return
    
    def AddThimbleAnalogRawSensor(self, analogRawSensorData: WeArtAnalogSensorData):
        return
        
    def SizeThimbles(self):
        return len(self.__thimbleTrackingObjects)
    
    def AddMessageListener(self, listener: WeArtMessageListener):
        self.__messageListeners.append(listener)

    def AddMessageCallback(self, callback):
        return

    def RemoveMessageListener(self, listener: WeArtMessageListener):
        if listener in self.__messageListeners:
            self.__messageListeners.remove(listener)

    def RemoveMessageCallback(self, callback):
        return

    def AddConnectionStatusCallback(self, callback):
        self.__connectionStatusCallbacks.append(callback)

    def AddErrorCallback(self, callback):
        self.__errorCallbacks.append(callback)
    
    def _OnReceive(self):
        try:
            while True:
                data = self.__s.recv(4096)
                str_data = data.decode()
                self.__logger.debug(f"Received: { str_data }")
                strings = str_data.split(WeArtClient.messagesSeparator)
                messages = []
                for string in strings:
                    messages.append(self._messageSerializer.Deserialize(string))
                self.__ForwardingMessages(messages)
        except Exception:
            if not self.__Closing:
                raise
    
    def __ForwardingMessages(self, messages:list[WeArtMessages.WeArtMessage]):
        for msg in messages:
            if msg == None:
                continue
            for listener in self.__messageListeners:
                if (listener.accept(msg.getID())):
                    listener.OnMessageReceived(msg)
    
    def __NotifyConnectionStatus(self, connected:bool):
        for callback in self.__connectionStatusCallbacks:
            t = Thread(target=callback, args = [connected])
            t.start()
            self.__pendingCallbacks.append(t)
        self.__ClearProcessedCallbacks()
    
    def __NotifyError(self, errorType:ErrorType):
        for callback in self.__errorCallbacks:
            t = Thread(target=callback, args = [errorType])
            t.start()
            self.__pendingCallbacks.append(t)
        self.__ClearProcessedCallbacks()
    
    def __ClearProcessedCallbacks(self):
        for t in self.__pendingCallbacks:
            t.join()

__all__ = ['WeArtClient']


from .WeArtCommon import TextureType

class WeArtTexture:
    DefaultTextureType = TextureType.ClickNormal
    
    DefaultVolume = 100.0
    MinVolume = 0.0
    MaxVolume = 100.0

    DefaultVelocity = 0.0
    MinVelocity = 0.0
    MaxVelocity = 0.5

    def __init__(self, active = False, texture_type = TextureType.ClickNormal, velocity = 0.0, volume = 100.0):
        self.active = active
        self.__textureType = texture_type
        self.__textureVelocity = velocity
        self.__volume = volume
    
    @property
    def textureVelocity(self):
        return self.__textureVelocity
    
    @textureVelocity.setter
    def textureVelocity(self, velocity:float):
        if velocity < self.MinVelocity:
            self.__textureVelocity = self.MinVelocity
        elif velocity > self.MaxVelocity:
            self.__textureVelocity = self.MaxVelocity
        else:
            self.__textureVelocity = velocity

    @property
    def textureType(self):
        return self.__textureType
    
    @textureType.setter
    def textureType(self, texture_type:TextureType):
        lower = TextureType.ClickNormal
        upper = TextureType.DoubleSidedTape

        if texture_type.value <= lower.value:
            self.__textureType = lower
        elif texture_type.value >= upper.value:
            self.__textureType = upper
        else:
            self.__textureType = texture_type
    
    @property
    def volume(self):
        return self.__volume
    
    @volume.setter
    def volume(self, volume:float):
        if volume < self.MinVolume:
            self.__volume = self.MinVolume
        elif volume > self.MaxVolume:
            self.__volume = self.MaxVolume
        else:
            self.__volume = volume

    def __eq__(self, other):
        if (self.active == other.active and self.__textureType == other.textureType and self.__textureVelocity == other.textureVelocity and self.__volume == other.volume):
            return True
        else:
            return False
    
__all__ = ['WeArtTexture']



from .WeArtCommon import TrackingType, HandSide, ActuationPoint, CalibrationStatus, SensorData, AnalogSensorRawData, MiddlewareStatusData, ConnectedDeviceStatus
from .WeArtCommon import dataclass_from_dict, dict_from_dataclass, dataclass_from_list
from . import WeArtCommon
import json
import time
import logging

def StringToTrackingType(string:str):
	if (string == "TrackType1"):
		return TrackingType.WEART_HAND
	return TrackingType.DEFAULT

def TrackingTypeToString(trackType:TrackingType):
    if trackType == TrackingType.DEFAULT:
        return ""
    elif trackType == TrackingType.WEART_HAND:
        return "TrackType1"
    else:
	    return ""
    
def StringToHandside(string:str):
    if (string == "LEFT"):
        return HandSide.Left
    elif (string == "RIGHT"):
        return HandSide.Right
    else:
        assert(False)
        return HandSide.Left

def HandsideToString(hs:HandSide):
    if (hs == HandSide.Left):
        return "LEFT"
    elif (hs == HandSide.Right):
        return "RIGHT"
    else:
        assert(False)
        return ""


def StringToActuationPoint(string:str):
    if string == "THUMB":
        return ActuationPoint.Thumb
    elif string == "INDEX":
        return ActuationPoint.Index
    elif string == "MIDDLE":
        return ActuationPoint.Middle
    elif string == "PALM":
        return ActuationPoint.Palm
    else:
        assert(False)
        return ActuationPoint.Thumb


def ActuationPointToString(ap: ActuationPoint):
    if ap == ActuationPoint.Thumb:
        return "THUMB"
    elif ap == ActuationPoint.Index:
        return "INDEX"
    elif ap == ActuationPoint.Middle:
        return "MIDDLE"
    elif ap == ActuationPoint.Palm:
        return "PALM"
    else:
        assert(False)
        return ""
    
def CalibrationHandSideToString(hs: HandSide):
    return "0" if hs == HandSide.Left else "1"
    

def StringToCalibrationHandSide(string: str):
    if (string == "0"):
        return HandSide.Left
    elif (string == "1"):
        return HandSide.Right
    assert(False)
    return HandSide.Left


class WeArtMessage:
	#@brief Allows to get the message ID, used to deserialize the correct message type
	#@return the message ID
    def __init__(self):
        return

    def getID(self):
        return

    def getValues(self):
        return []

    def setValues(self):
        return

    def setHandSide(self, handside:HandSide):
        return

    def setActuationPoint(self, actuation_point:ActuationPoint):
        return

class WeArtCsvMessage(WeArtMessage):
    field_separator = ':'
    def __init__(self):
        super().__init__()

    def getValues(self):
        return super().getValues()
    
    def setValues(self):
        return super().setValues()
    
    def serialize(self)->str:
        messageID = self.getID()
        serializedValues = self.getValues()

        serializedValues.insert(0, messageID)

        ss = serializedValues[0]
        for i in range(1, len(serializedValues)):
            ss += self.field_separator
            ss += serializedValues[i]

        return ss
    
    def deserialize(self, message:str)->None:
        strings = message.split(":")
        strings = strings[1:]
        self.setValues(strings)
        return 

class WeArtJsonMessage(WeArtMessage):
    def __init__(self):
        super().__init__()
        self._timestamp = int(time.time() * 1000)
    
    def timestamp(self):
        return self._timestamp
    
    def serialize(self) -> str:
        j = {}
        j["type"] = self.getID()
        j["ts"] = self._timestamp
        payload = self._serializePayload()
        if payload != None:
            j["data"] = payload
        
        return json.dumps(j)
    
    def deserialize(self, message:str)->None:
        j = json.loads(message)
        self._timestamp = int(j["ts"])
        if "data" in j:
            self._deserializePayload(j["data"])
    
    def setHandSide(self, handside: HandSide)->None:
        return 
    
    def setActuationPoint(self, actuation_point: ActuationPoint)->None:
        return 

    def _serializePayload(self):
        return
    
    def _deserializePayload(self):
        return
    

class WeArtMessageNoParams(WeArtCsvMessage):
    def __init__(self):
        super().__init__()
    # the message carries no value, so return empty list
    def getValues(self):
        return []
    # the message carries no value, so return without doint anything
    def setValues(self, values):
        return
    


class WeArtMessageObjectSpecific(WeArtCsvMessage):
    def __init__(self):
        self._handSide = None
        self._actuationPoint = None
    
    def setHandSide(self, handside:HandSide):
        self._handSide = handside
    
    def getActuationPoint(self):
        return self._actuationPoint
    
    def setActuationPoint(self, actuation_point:ActuationPoint):
        self._actuationPoint = actuation_point


class WeArtMessageHandSpecific(WeArtCsvMessage):
    def __init__(self):
        super().__init__()
        self._handSide = None

    def getHand(self):
        return self._handSide

    def setHandSide(self, handside: HandSide):
        self._handSide = handside
    
    def setActuationPoint(self, actuation_point: ActuationPoint):
        return


class StartFromClientMessage(WeArtMessageNoParams):
    ID = "StartFromClient"

    def __init__(self, trackType = TrackingType.WEART_HAND):
        self._tracktype = trackType

    def getID(self):
        return self.ID
    
    def getValues(self):
        res = []
        if self._tracktype != TrackingType.DEFAULT:
            res.append(WeArtCommon.WEART_SDK_TYPE)
            res.append(WeArtCommon.WEART_SDK_VERSION)
            res.append(TrackingTypeToString(self._tracktype))
        return res

    def setValues(self, values:list[str]):
        if (len(values) == 0):
            self._trackType = TrackingType.DEFAULT
        else:
            self._trackType = StringToTrackingType(values[2])
        return

class StopFromClientMessage(WeArtMessageNoParams):
    ID = "StopFromClient"

    def getID(self):
        return self.ID
    

class StartCalibrationMessage(WeArtMessageNoParams):
	ID = "StartCalibration"

	def getID(self): 
		return self.ID

class StopCalibrationMessage(WeArtMessageNoParams):
	ID = "StopCalibration"

	def getID(self): 
		return self.ID

class CalibrationStatusMessage(WeArtMessageHandSpecific):
    ID = "CalibrationStatus"

    def __init__(self):
        super().__init__()
        self._status = None

    def getID(self):
        return self.ID
    
    def getStatus(self):
        return self._status
    
    def getValues(self):
        ret = []
        ret.append(CalibrationHandSideToString(self._handSide))
        ret.append(str(int(self._status)))
        return ret

    def setValues(self, values:list[str]):
        assert(len(values) == 2)
        self._handSide = StringToCalibrationHandSide(values[0])
        self._status = CalibrationStatus(int(values[1]))



class CalibrationResultMessage(WeArtMessageHandSpecific):
    ID = "CalibrationResult"

    def __init__(self):
        super().__init__()
        self._success = False

    def getID(self):
        return self.ID
    
    def getSuccess(self):
        return self._success
    
    def getValues(self):
        ret = []
        ret.append(CalibrationHandSideToString(self._handSide))
        ret.append("0" if self._success else "1")
        return ret 

    def setValues(self, values: list[str]):
        assert(len(values) == 2)
        self._handSide = StringToCalibrationHandSide(values[0])
        self._success = int(values[1]) == 0


class TrackingMessage(WeArtMessageNoParams):
    ID = "Tracking"

    def __init__(self, trackType = TrackingType.WEART_HAND):
        self.Angles = {"X":0.0, "Y": 0.0, "Z": 0.0}
        self._trackingType = trackType
        # Closures
        self.__RightThumbClosure = 0
        self.__RightIndexClosure = 0
        self.__RightMiddleClosure = 0
        self.__RightPalmClosure = 0
        self.__LeftThumbClosure = 0
        self.__LeftIndexClosure = 0
        self.__LeftMiddleClosure = 0
        self.__LeftPalmClosure = 0

        # Abductions
        self.__RightThumbAbduction = 0
        self.__LeftThumbAbduction = 0

    def getID(self):
        return self.ID

    def getValues(self):
        ret = []
        ret.append(TrackingTypeToString(self._trackingType))
        if self._trackingType == TrackingType.DEFAULT:
            ret.append(str(self.__RightThumbClosure))
            ret.append(str(self.__RightIndexClosure))
            ret.append(str(self.__RightMiddleClosure))
            ret.append(str(self.__RightPalmClosure))
            ret.append(str(self.__LeftThumbClosure))
            ret.append(str(self.__LeftIndexClosure))
            ret.append(str(self.__LeftMiddleClosure))
            ret.append(str(self.__LeftPalmClosure))
        elif self._trackingType == TrackingType.WEART_HAND:
            ret.append(str(self.__RightIndexClosure))
            ret.append(str(self.__RightThumbClosure))
            ret.append(str(self.__RightThumbAbduction))
            ret.append(str(self.__RightMiddleClosure))
            
            ret.append(str(self.__LeftIndexClosure))
            ret.append(str(self.__LeftThumbClosure))
            ret.append(str(self.__LeftThumbAbduction))
            ret.append(str(self.__LeftMiddleClosure))
        return ret
    
    def setValues(self, values:list[str]):
        self._trackingType = StringToTrackingType(values[0])
        if self._trackingType == TrackingType.DEFAULT:
            assert(len(values)==8)
            self.__RightThumbClosure = int(values[0])
            self.__RightIndexClosure = int(values[1])
            self.__RightMiddleClosure = int(values[2])
            self.__RightPalmClosure = int(values[3])
            self.__LeftThumbClosure = int(values[4])
            self.__LeftIndexClosure = int(values[5])
            self.__LeftMiddleClosure = int(values[6])
            self.__LeftPalmClosure = int(values[7])
        elif self._trackingType == TrackingType.WEART_HAND:
            assert(len(values)==9)
            self.__RightIndexClosure = int(values[1])
            self.__RightThumbClosure = int(values[2])
            self.__RightThumbAbduction = int(values[3])
            self.__RightMiddleClosure = int(values[4])

            self.__LeftIndexClosure = int(values[5])
            self.__LeftThumbClosure = int(values[6])
            self.__LeftThumbAbduction = int(values[7])
            self.__LeftMiddleClosure = int(values[8])

    def GetType(self):
        return self._trackingType
    
    def GetAbduction(self, handSide:HandSide, actuationPoint:ActuationPoint):
        maxAbductionValue = 255.0
        if handSide == HandSide.Left:
            if (actuationPoint == ActuationPoint.Thumb):
                return float(self.__LeftThumbAbduction) / maxAbductionValue
        elif handSide == HandSide.Right:
            if (actuationPoint == ActuationPoint.Thumb):
                return float(self.__RightThumbAbduction) / maxAbductionValue
        return WeArtCommon.defaultAbduction
    
    
    def GetClosure(self, handSide:HandSide, actuationPoint:ActuationPoint):
        byteValue = 0
        if handSide == HandSide.Left:
            if actuationPoint == ActuationPoint.Thumb:
                byteValue = self.__LeftThumbClosure
            elif actuationPoint == ActuationPoint.Index:
                byteValue = self.__LeftIndexClosure
            elif actuationPoint == ActuationPoint.Middle:
                byteValue = self.__LeftMiddleClosure
            elif actuationPoint == ActuationPoint.Palm:
                byteValue = self.__LeftPalmClosure
        elif handSide == HandSide.Right:
            if actuationPoint == ActuationPoint.Thumb:
                byteValue = self.__RightThumbClosure
            elif actuationPoint == ActuationPoint.Index:
                byteValue = self.__RightIndexClosure
            elif actuationPoint == ActuationPoint.Middle:
                byteValue = self.__RightMiddleClosure
            elif actuationPoint == ActuationPoint.Palm:
                byteValue = self.__RightPalmClosure
        closure = float(byteValue) / float(255)
        return closure
    
class SetTemperatureMessage(WeArtMessageObjectSpecific):
    ID = "temperature"

    def __init__(self, t:float):
        super().__init__()
        self._temperature = t
    
    def getID(self):
        return self.ID
    
    def getValues(self):
        ret = []
        ret.append(str(self._temperature))
        ret.append(HandsideToString(self._handSide))
        ret.append(ActuationPointToString(self._actuationPoint))
        return ret
    
    def setValues(self, values:list[str]):
        assert(len(values)==3)
        self._temperature = float(values[0])
        self._handSide = StringToHandside(values[1])
        self._actuationPoint= StringToActuationPoint(values[2])

    
class StopTemperatureMessage(WeArtMessageObjectSpecific):
    ID = "stopTemperature"

    def __init__(self):
        super().__init__()

    def getID(self):
        return self.ID
    

    def getValues(self):
        ret = [HandsideToString(self._handSide), ActuationPointToString(self._actuationPoint)]
        return ret
    
    def setValues(self, values: list[str]):
        assert(len(values) == 2)
        self._handSide = StringToHandside(values[0])
        self._actuationPoint = StringToActuationPoint(values[1])

class SetForceMessage(WeArtMessageObjectSpecific):
    ID = "force"

    def __init__(self, force:list[float]):
        super().__init__()
        self._force = force

    def getID(self):
        return self.ID
    
    def getValues(self):
        ret = []
        ret.append(str(self._force[0]))
        ret.append(str(self._force[1]))
        ret.append(str(self._force[2]))
        ret.append(HandsideToString(self._handSide))
        ret.append(ActuationPointToString(self._actuationPoint))
        return ret
    
    def setValues(self, values:list[str]):
        assert(len(values) == 5)
        self._force[0] = float(values[0])
        self._force[1] = float(values[1])
        self._force[2] = float(values[2])
        self._handSide = StringToHandside(values[3])
        self._actuationPoint = StringToActuationPoint(values[4])

class StopForceMessage(WeArtMessageObjectSpecific):
    ID = "stopForce"

    def __init__(self):
        super().__init__()

    def getID(self):
        return self.ID
    

    def getValues(self):
        ret = [HandsideToString(self._handSide), ActuationPointToString(self._actuationPoint)]
        return ret
    
    def setValues(self, values: list[str]):
        assert(len(values) == 2)
        self._handSide = StringToHandside(values[0])
        self._actuationPoint = StringToActuationPoint(values[1])

class SetTextureMessage(WeArtMessageObjectSpecific):
    ID = "texture"

    def __init__(self, idx:int, vel:float, vol:float):
        super().__init__()
        self._index = idx
        self._velocity = [0.5, 0.0, vel]
        self._volume = vol
    
    def getID(self):
        return self.ID

    def getValues(self):
        ret = []
        if (self._index < WeArtCommon.minTextureIndex or self._index > WeArtCommon.maxTextureIndex):
            self._index = WeArtCommon.nullTextureIndex
        
        ret.append(str(self._index))
        ret.append(str(self._velocity[0]))
        ret.append(str(self._velocity[1]))
        ret.append(str(self._velocity[2]))
        ret.append(str(self._volume))
        ret.append(HandsideToString(self._handSide))
        ret.append(ActuationPointToString(self._actuationPoint))
        return ret

    def setValues(self, values:list[str]):
        assert(len(values) == 6)
        self._index = int(values[0])
        self._velocity = []
        self._velocity.append(float(values[1]))
        self._velocity.append(float(values[2]))
        self._velocity.append(float(values[3]))
        self._volume = float(values[4])
        self._handSide = StringToHandside(values[5])
        self._actuationPoint = StringToActuationPoint(values[6])


class StopTextureMessage(WeArtMessageObjectSpecific):
    ID = "stopTexture"

    def __init__(self):
        super().__init__()

    def getID(self):
        return self.ID
    

    def getValues(self):
        ret = [HandsideToString(self._handSide), ActuationPointToString(self._actuationPoint)]
        return ret
    
    def setValues(self, values: list[str]):
        assert(len(values) == 2)
        self._handSide = StringToHandside(values[0])
        self._actuationPoint = StringToActuationPoint(values[1])


class RawDataOn(WeArtJsonMessage):
    ID = "RAW_DATA_ON"  
    def __init__(self):
        super().__init__()
    
    def getID(self):
        return self.ID

    def setHandSide(self, handside: HandSide) -> None:
        return super().setHandSide(handside)
    
    def setActuationPoint(self, actuation_point: ActuationPoint) -> None:
        return super().setActuationPoint(actuation_point)


class RawDataOff(WeArtJsonMessage):
    ID = "RAW_DATA_OFF"  
    def __init__(self):
        super().__init__()
    
    def getID(self):
        return self.ID

    def setHandSide(self, handside: HandSide) -> None:
        return super().setHandSide(handside)
    
    def setActuationPoint(self, actuation_point: ActuationPoint) -> None:
        return super().setActuationPoint(actuation_point)
    
class RawSensorsData(WeArtJsonMessage):
    ID = "RAW_DATA"
    def __init__(self):
        super().__init__()
        self._hand = None
        self._sensors = {} #std::map<ActuationPoint, SensorData> sensors;

    def getID(self):
        return self.ID
    
    def setHandSide(self, handside: HandSide) -> None:
        self._hand = handside

    def setActuationPoint(self, actuation_point: ActuationPoint) -> None:
        return
    
    def getHand(self)->HandSide:
        return self._hand
    
    def hasSensor(self, ap:ActuationPoint) -> bool:
         return (ap in self._sensors)
        
    
    def getSensor(self, ap:ActuationPoint) -> SensorData:
       return self._sensors[ap]
    
    def _serializePayload(self):
        j = {}
        j["handSide"] = self._hand
        for s in self._sensors:
            actuationPoint = ActuationPointToString(s).lower()
            j[actuationPoint] = dict_from_dataclass(self._sensors[s])
        return j
    
    def _deserializePayload(self, payload):
        hs = payload["handSide"]
        self._hand = StringToHandside(hs)
        if "index" in payload:
            self._sensors[ActuationPoint.Index] = dataclass_from_dict(SensorData, payload["index"])
        if "thumb" in payload:
            self._sensors[ActuationPoint.Thumb] = dataclass_from_dict(SensorData, payload["thumb"])
        if "middle" in payload:
            self._sensors[ActuationPoint.Middle] = dataclass_from_dict(SensorData, payload["middle"])
        if "palm" in payload:
            self._sensors[ActuationPoint.Palm] = dataclass_from_dict(SensorData, payload["palm"])


class AnalogSensorsData(WeArtJsonMessage):
    ID = "RAW_SENSOR_ON_MASK"
    def __init__(self):
        super().__init__()
        self._hand = None
        self._sensors = {} #std::map<ActuationPoint, AnalogSensorRawData> sensors;
    
    def getID(self):
        return self.ID
    
    def setHandSide(self, handside: HandSide) -> None:
        self._hand = handside

    def setActuationPoint(self, actuation_point: ActuationPoint) -> None:
        return

    def getHand(self) -> HandSide:
        return self._hand
    
    def hasSensor(self, ap:ActuationPoint) -> bool:
        return ap in self._sensors
    
    def getSensor(self, ap:ActuationPoint) -> AnalogSensorRawData:
        return self._sensors[ap]
    
    def _serializePayload(self):
        j = {}
        j["handSide"] = self._hand
        for s in self._sensors:
            actuationPoint = ActuationPointToString(s).lower()
            j[actuationPoint] = dict_from_dataclass(self._sensors[s])
        return j
    
    def _deserializePayload(self, payload) -> None:
        hs = payload["handSide"]
        self._hand = StringToHandside(hs)
        if "index" in payload:
            self._sensors[ActuationPoint.Index] = dataclass_from_dict(AnalogSensorRawData, payload["index"])
        if "thumb" in payload:
            self._sensors[ActuationPoint.Thumb] = dataclass_from_dict(AnalogSensorRawData, payload["thumb"])
        if "middle" in payload:
            self._sensors[ActuationPoint.Middle] = dataclass_from_dict(AnalogSensorRawData, payload["middle"])
        if "palm" in payload:
            self._sensors[ActuationPoint.Palm] = dataclass_from_dict(AnalogSensorRawData, payload["palm"])

class GetMiddlewareStatus(WeArtJsonMessage):

    ID = "MW_GET_STATUS"
    def __init__(self):
        super().__init__()

    def getID(self):
        return self.ID
    
class MiddlewareStatusMessage(WeArtJsonMessage):
    
    ID = "MW_STATUS"

    def __init__(self):
        super().__init__()
        self.__data = None #MiddlewareStatusData _data;

    def getID(self) -> str:
        return self.ID

    def data(self) -> MiddlewareStatusData:
        return self.__data
    
    def _serializePayload(self) -> dict:
        return dict_from_dataclass(self.__data)
    
    def _deserializePayload(self, payload:dict) -> None:
        self.__data = dataclass_from_dict(MiddlewareStatusData, payload)
        try:
            self.__data.timestamp = self._timestamp
        except:
            pass
        

class GetDevicesStatusMessage(WeArtJsonMessage):

    ID = "DEVICES_GET_STATUS"
    def __init__(self):
        super().__init__()
    
    def getID(self):
        return self.ID
    
class DevicesStatusMessage(WeArtJsonMessage):
    ID = "DEVICES_STATUS"

    def __init__(self):
        super().__init__()
        self.__devices = [] #std::vector<ConnectedDeviceStatus> _devices;

    def getID(self):
        return self.ID
    
    def devices(self) -> list:
        return self.__devices
    
    def _serializePayload(self) -> dict:
        j = {}
        j["devices"] = dict_from_dataclass(self.__devices)
        return j
    
    def _deserializePayload(self, payload:dict) -> None:
        if "devices" in payload:
            self.__devices = dataclass_from_list(ConnectedDeviceStatus, payload["devices"])








    


  






from .WeArtMessageListener import WeArtMessageListener
from .WeArtCommon import HandSide, ActuationPoint, SensorData
from .WeArtMessages import WeArtMessage, RawSensorsData

from dataclasses import dataclass

@dataclass
class Sample:
    timestamp: int
    data: SensorData

class WeArtTrackingRawData(WeArtMessageListener):
    def __init__(self, handSide:HandSide, actuationPoint:ActuationPoint):
        super().__init__(RawSensorsData.ID)
        self.__handSide = handSide
        self.__actuationPoint = actuationPoint
        self.__K_NUM_SAMPLES = 3
        self.__samples = [] #std::queue<Sample> samples;
        self.__callbacks =  [] #std::vector<std::function<void(Sample)>> callbacks;

    def GetLastSample(self) -> Sample:
        if len(self.__samples) == 0:
            return Sample(0, SensorData(0, 0, 0))
        return self.__samples[-1]

    def AddSampleCallback(self, callback):
        self.__callbacks.append(callback)
    
    def OnMessageReceived(self, message: WeArtMessage):
        if message.getID() != RawSensorsData.ID: #to be checked
            return
        rawSensorsData = message
        if rawSensorsData == None:
            return
        if rawSensorsData.getHand() != self.__handSide:
            return
        if not rawSensorsData.hasSensor(self.__actuationPoint):
            return 

        sample = Sample(timestamp = rawSensorsData.timestamp(), data = rawSensorsData.getSensor(self.__actuationPoint))
        self.__samples.append(sample)

        while len(self.__samples) > self.__K_NUM_SAMPLES:
            self.__samples.pop(0)
        
        for callback in self.__callbacks:
            callback(sample)

__all__ = ['WeArtTrackingRawData']



from .WeArtMessageListener import WeArtMessageListener
from .WeArtMessages import CalibrationResultMessage, CalibrationStatusMessage, WeArtMessage

class WeArtTrackingCalibration(WeArtMessageListener):
    def __init__(self):
        super().__init__([CalibrationStatusMessage.ID, CalibrationResultMessage.ID])
        self.__currentHand = None
        self.__status = None
        self.__result = False
        self.__statusCallbacks = []
        self.__resultCallbacks = []

    def getCurrentHand(self):
        return self.__currentHand
    
    def getStatus(self):
        return self.__status
    
    def getResult(self):
        return self.__result
    
    def AddStatusCallback(self, callback):
        self.__statusCallbacks.append(callback)
    
    def AddResultCallback(self, callback):
        self.__resultCallbacks.append(callback)
    
    def OnMessageReceived(self, message:WeArtMessage):
        if message.getID() == CalibrationStatusMessage.ID:
            self.__currentHand = message.getHand()
            self.__status = message.getStatus()
            for callback in self.__statusCallbacks:
                callback(self.__currentHand, self.__status)
        elif message.getID() == CalibrationResultMessage.ID:
            self.__currentHand = message.getHand()
            self.__result = message.getSuccess()
            for callback in self.__resultCallbacks:
                callback(self.__currentHand, self.__result)

__all__ = ['WeArtTrackingCalibration']



class WeArtTemperature:
    DefaultValue = 0.5
    MinValue = 0.0
    MaxValue = 1.0

    def __init__(self):
        self.active = False
        self.__value = self.DefaultValue
    
    @property
    def value(self):
        return self.__value
    
    @value.setter
    def value(self, temperature:float):
        if temperature < self.MinValue:
            self.__value = self.MinValue
        elif temperature > self.MaxValue:
            self.__value = self.MaxValue
        else:
            self.__value = temperature

    def __eq__(self, other):
        if (self.active == other.active and self.__value == other.value):
            return True
        else:
            return False

__all__ = ['WeArtTemperature']



from . import WeArtMessages as WeArtMessages

class WeArtMessageListener:
    def __init__(self, ids:list[str]):
        self._acceptedIds = ids

    def accept(self, id:str):
        if id in self._acceptedIds:
            return True
        else:
            return False
    def OnMessageReceived(message:WeArtMessages.WeArtMessage):
        return



from .WeArtMessageListener import WeArtMessageListener
from .WeArtCommon import HandSide, ActuationPoint, AnalogSensorRawData
from .WeArtMessages import WeArtMessage, AnalogSensorsData
from dataclasses import dataclass


@dataclass
class Sample:
    timestamp: int = 0
    data: AnalogSensorRawData = None

class WeArtAnalogSensorData(WeArtMessageListener):

    def __init__(self, handSide:HandSide, actuationPoint:ActuationPoint):
        super().__init__([AnalogSensorsData.ID])
        self.__lastSample = Sample()
        self.__callbacks = [] #std::vector<std::function<void(Sample)>> callbacks;
        self.__handSide = handSide
        self.__actuationPoint = actuationPoint


    def GetLastSample(self) -> Sample:
        return self.__lastSample
    
    def AddSampleCallback(self, callback):
        self.__callbacks.append(callback)
    
    def OnMessageReceived(self, message: WeArtMessage):
        if message.getID() != AnalogSensorsData.ID: #to be checked
            return 
        rawSensorsData = message # cast to AnalogSensorsData
        if rawSensorsData == None:
            return
        if rawSensorsData.getHand() != self.__handSide:
            return
        if not rawSensorsData.hasSensor(self.__actuationPoint):
            return
        
        sample = Sample(timestamp = rawSensorsData.timestamp(), data = rawSensorsData.getSensor(self.__actuationPoint))
        self.__lastSample = sample
        for callback in self.__callbacks:
            callback(sample)

__all__ = ['WeArtAnalogSensorData']


from .WeArtTemperature import WeArtTemperature
from .WeArtForce import WeArtForce
from .WeArtTexture import WeArtTexture

class WeArtEffect:
    def __init__(self):
        return
    
    def getTemperature():
        return
    def getForce():
        return
    def getTexture():
        return

class TouchEffect(WeArtEffect):
    def __init__(self, temp:WeArtTemperature, force:WeArtForce, texture:WeArtTexture):
        self.__temperature = temp
        self.__force = force
        self.__texture = texture

    def getTemperature(self):
        return self.__temperature

    def getForce(self):
        return self.__force
    
    def getTexture(self):
        return self.__texture
    
    def Set(self, temperature:WeArtTemperature, force:WeArtForce, texture = WeArtTexture):
        changed = False

        if temperature == self.__temperature:
            changed = True
        self.__temperature = temperature

        if force == self.__force:
            changed = True
        self.__force = force

        if texture == self.__texture:
            changed = True
        texture.textureVelocity = 0.5
        self.__texture = texture

        return changed

__all__ = ['WeArtEffect', 'TouchEffect']



from . import WeArtMessages as WeArtMessages
import json

class WeArtMessageSerializer:
    separator = ':'
    
    @staticmethod
    def HashStringToInt(string:str, _hash = 0):
        return WeArtMessageSerializer._hashstringtoint(string, 0, _hash)
    
    @staticmethod
    def _hashstringtoint(string:str, index:int, _hash = 0):
        if (index == len(string)):
            return _hash
        else:
            return 101 * WeArtMessageSerializer._hashstringtoint(string, index+1, _hash) + ord(string[index])


	# @brief Creates a message of the right type given its type ID
	# @param ID Type of the message to create
	# @return the created message
        
    def __createMessage(self, id:str):
        h = WeArtMessageSerializer.HashStringToInt(id)
        if (h == WeArtMessageSerializer.HashStringToInt(WeArtMessages.StartFromClientMessage.ID)):
            return WeArtMessages.StartFromClientMessage()
        elif (h == WeArtMessageSerializer.HashStringToInt(WeArtMessages.TrackingMessage.ID)):
            return WeArtMessages.TrackingMessage()
        elif (h == WeArtMessageSerializer.HashStringToInt(WeArtMessages.StopFromClientMessage.ID)):
            return WeArtMessages.StopFromClientMessage()
        elif (h == WeArtMessageSerializer.HashStringToInt(WeArtMessages.CalibrationResultMessage.ID)):
            return WeArtMessages.CalibrationResultMessage()
        elif (h == WeArtMessageSerializer.HashStringToInt(WeArtMessages.CalibrationStatusMessage.ID)):
            return WeArtMessages.CalibrationStatusMessage()
        elif (h == WeArtMessageSerializer.HashStringToInt(WeArtMessages.SetTemperatureMessage.ID)):
            return WeArtMessages.SetTemperatureMessage()
        elif (h == WeArtMessageSerializer.HashStringToInt(WeArtMessages.StopTemperatureMessage.ID)):
            return WeArtMessages.StopTemperatureMessage()
        elif (h == WeArtMessageSerializer.HashStringToInt(WeArtMessages.SetForceMessage.ID)):
            return WeArtMessages.SetForceMessage()
        elif (h == WeArtMessageSerializer.HashStringToInt(WeArtMessages.StopForceMessage.ID)):
            return WeArtMessages.StopForceMessage()
        elif (h == WeArtMessageSerializer.HashStringToInt(WeArtMessages.SetTextureMessage.ID)):
            return WeArtMessages.SetTextureMessage()
        elif (h == WeArtMessageSerializer.HashStringToInt(WeArtMessages.StopTextureMessage.ID)):
            return WeArtMessages.StopTextureMessage()
        elif (h == WeArtMessageSerializer.HashStringToInt(WeArtMessages.RawSensorsData.ID)):
            return WeArtMessages.RawSensorsData()
        elif (h == WeArtMessageSerializer.HashStringToInt(WeArtMessages.GetMiddlewareStatus.ID)):
            return WeArtMessages.GetMiddlewareStatus()
        elif (h == WeArtMessageSerializer.HashStringToInt(WeArtMessages.MiddlewareStatusMessage.ID)):
            return WeArtMessages.MiddlewareStatusMessage()
        elif (h == WeArtMessageSerializer.HashStringToInt(WeArtMessages.GetDevicesStatusMessage.ID)):
            return WeArtMessages.GetDevicesStatusMessage()
        elif (h == WeArtMessageSerializer.HashStringToInt(WeArtMessages.DevicesStatusMessage.ID)):
            return WeArtMessages.DevicesStatusMessage()
        elif (h == WeArtMessageSerializer.HashStringToInt(WeArtMessages.AnalogSensorsData.ID)):
            return WeArtMessages.AnalogSensorsData()
        #TODO Continues...
        else: 
            return None
        '''   
            case HashStringToInt(ExitMessage::ID): return new ExitMessage();
            case HashStringToInt(DisconnectMessage::ID): return new DisconnectMessage();
        '''

    def Deserialize(self, data: str) -> WeArtMessages.WeArtMessage:
        id = self.__extractID(data)
        msg = self.__createMessage(id)
        if msg != None:
            msg.deserialize(data)
        return msg

    def __extractID(self, data:str) -> str:
        try:
            # if it is a JSON message
            d = json.loads(data)
            return d["type"]
        except:
            # if it is a CSV message
            return data.split(":")[0]

	# Bytestream to std::string.

	# @brief Converts a bytestream to string
	# @param byteData Pointer to the data buffer to convert
	# @param byteCount Size of the data buffer to convert
	# @return deserializes buffer as string
	#std::string Deserialize(uint8* byteData, int byteCount);

    #TODO: maybe to be implemented



# internal, not wildcarded: WeArtMessageListener, WeArtMessages, WeArtMessageSerializer

from . import MiddlewareStatusListener, WeArtAnalogSensorData, WeArtClient, WeArtCommon, WeArtEffect, WeArtForce, WeArtHapticObject, WeArtTemperature, WeArtTexture, WeArtThimbleTrackingObject, WeArtTrackingCalibration, WeArtTrackingRawData

__all__ = ['WeArtCommon']
for submod in (MiddlewareStatusListener, WeArtAnalogSensorData, WeArtClient, WeArtCommon, WeArtEffect, WeArtForce, WeArtHapticObject, WeArtTemperature, WeArtTexture, WeArtThimbleTrackingObject, WeArtTrackingCalibration, WeArtTrackingRawData):
    __all__.extend(submod.__all__)

from .MiddlewareStatusListener import *
from .WeArtAnalogSensorData import *
from .WeArtClient import *
# from .WeArtCommon import * # we have access to all objects through WeArtCommon.<something>
from .WeArtEffect import *
from .WeArtForce import *
from .WeArtHapticObject import *
from .WeArtTemperature import *
from .WeArtTexture import *
from .WeArtThimbleTrackingObject import *
from .WeArtTrackingCalibration import *
from .WeArtTrackingRawData import *



from .WeArtClient import WeArtClient
from .WeArtMessages import WeArtMessage, StopTemperatureMessage, StopForceMessage, StopTextureMessage
from .WeArtMessages import SetTemperatureMessage, SetForceMessage, SetTextureMessage
from .WeArtEffect import WeArtEffect
from .WeArtCommon import HandSide, ActuationPoint
from . import WeArtCommon
from .WeArtForce import WeArtForce
from .WeArtTemperature import WeArtTemperature
from .WeArtTexture import WeArtTexture

class WeArtHapticObject:
    def __init__(self, client:WeArtClient):
        self.handSideFlag = None
        self.actuationPointFlag = None
        self.weArtTemperature = None
        self.weArtForce = None
        self.weArtTexture = None
        self.activeEffects = []
        self.__client = client

    def AddEffect(self, effect:WeArtEffect):
        if not self.ContainsEffect(effect):
            self.activeEffects.append(effect)
            self.UpdateEffects()

    def RemoveEffect(self, effect:WeArtEffect):
        if self.ContainsEffect(effect):
            self.activeEffects.remove(effect)
            self.UpdateEffects()

    def ContainsEffect(self, effect:WeArtEffect):
        if effect in self.activeEffects:
            return True
        else:
            return False
        
    def UpdateEffects(self):
        if len(self.activeEffects) == 0:
            self.weArtForce = WeArtForce(active = False, force = WeArtCommon.defaultForce)
            self.weArtTemperature = WeArtTemperature()
            self.weArtTemperature.active = False
            self.weArtTemperature.value(WeArtCommon.defaultTemperature)
            self.weArtTexture = WeArtTexture(active = False, texture_type=WeArtCommon.defaultTextureIndex, velocity = WeArtTexture.DefaultVelocity)
            msg1 = StopTemperatureMessage()
            self.SendMessage(msg1)
            msg2 = StopForceMessage()
            self.SendMessage(msg2)
            msg3 = StopTextureMessage()
            self.SendMessage(msg3)
        else:
            tempEffect = None
            for effect in self.activeEffects:
                if effect.getTemperature().active:
                    tempEffect = effect
                    break
            
            # update temperature
            newTemp = None
            if tempEffect is None:
                newTemp = WeArtTemperature()
            else:
                newTemp = tempEffect.getTemperature()
            if self.weArtTemperature is None or (not newTemp == self.weArtTemperature):
                if not newTemp.active:
                    msg = StopTemperatureMessage()
                    self.SendMessage(msg)
                else:
                    msg = SetTemperatureMessage(newTemp.value)
                    self.SendMessage(msg)
                self.weArtTemperature = newTemp
            
            # update force
            forceEffect = None
            for effect in self.activeEffects:
                if effect.getForce().active:
                    forceEffect = effect
                    break
            newForce = None
            if forceEffect is not None:
                newForce = forceEffect.getForce()
            else:
                newForce = WeArtForce()
            if self.weArtForce is None or (not (newForce == self.weArtForce)):
                if not (newForce.active):
                    msg = StopForceMessage()
                    self.SendMessage(msg)
                else:
                    fValue = [newForce.value, 0.0, 0.0]
                    msg = SetForceMessage(fValue)
                    self.SendMessage(msg)
                self.weArtForce = newForce

            # update texture
            texEffect = None
            for effect in self.activeEffects:
                if effect.getTexture().active:
                    texEffect = effect
                    break
            newTex = None
            if texEffect is not None:
                newTex = texEffect.getTexture()
            else:
                newTex = WeArtTexture()
            if self.weArtTexture is None or (not (newTex == self.weArtTexture)):
                if not (newTex.active):
                    msg = StopTextureMessage()
                    self.SendMessage(msg)
                else:
                    msg = SetTextureMessage(newTex.textureType, newTex.textureVelocity, newTex.volume)
                    self.SendMessage(msg)
                self.weArtTexture = newTex

    
    def SendMessage(self, msg:WeArtMessage):
        hand_sides = [HandSide.Left, HandSide.Right]
        actuation_points = [ActuationPoint.Thumb, ActuationPoint.Index, ActuationPoint.Middle, ActuationPoint.Palm]
        for hs in hand_sides:
            if hs & self.handSideFlag:
                for ap in actuation_points:
                    if ap & self.actuationPointFlag:
                        msg.setHandSide(hs)
                        msg.setActuationPoint(ap)
                        self.__client.SendMessage(msg)

__all__ = ['WeArtHapticObject']



from dataclasses import dataclass, field
from typing import List
from .WeArtCommon import MiddlewareStatus, ConnectedDeviceStatus
from .WeArtMessageListener import WeArtMessageListener
from .WeArtMessages import MiddlewareStatusMessage, DevicesStatusMessage
from .WeArtMessages import WeArtMessage
import logging

@dataclass 
class MiddlewareStatusUpdate:
    timestamp: int = 0
    status: MiddlewareStatus = MiddlewareStatus(0)
    version: str = ""
    statusCode: int = 0
    errorDesc: str = ""
    actuationsEnabled:bool = False
    devices: List = field(default_factory=lambda: [])

class MiddlewareStatusListener(WeArtMessageListener):
    def __init__(self):
        super().__init__([MiddlewareStatusMessage.ID, DevicesStatusMessage.ID])
        self.__statusCallbacks = []
        self.__data = MiddlewareStatusUpdate()
        self.__logger = logging.getLogger("WeArtClient")
    
    def OnMessageReceived(self, message: WeArtMessage):
        toUpdate = False
        if message.getID() == MiddlewareStatusMessage.ID:
            mwStatus = message
            newStatus = mwStatus.data()
            self.__data.timestamp = mwStatus.timestamp()
            self.__data.status = newStatus.status
            self.__data.version = newStatus.version
            self.__data.statusCode = newStatus.statusCode
            self.__data.errorDesc = newStatus.errorDesc
            self.__data.actuationsEnabled = newStatus.actuationsEnabled
            self.__logger.debug(self.__data)
            toUpdate = True

        elif message.getID() == DevicesStatusMessage.ID:
            deviceStatus = message
            if deviceStatus!= None:
                self.__data.timestamp = deviceStatus.timestamp()
                self.__data.devices = deviceStatus.devices()
                self.__logger.debug(self.__data)
                toUpdate = True

        if toUpdate:
            for callback in self.__statusCallbacks:
                callback(self.__data)
                

    def AddStatusCallback(self, callback):
        self.__statusCallbacks.append(callback)
    
    def lastStatus(self) -> MiddlewareStatusUpdate:
        return self.__data

__all__ = ['MiddlewareStatusListener', 'MiddlewareStatusUpdate']


from .WeArtMessageListener import WeArtMessageListener
from .WeArtCommon import HandSide, ActuationPoint
from .WeArtMessages import WeArtMessage, TrackingMessage
from . import WeArtCommon


class WeArtThimbleTrackingObject(WeArtMessageListener):
    def __init__(self, handSide:HandSide, actuationPoint:ActuationPoint):
        super().__init__([TrackingMessage.ID])
        self._handSide = handSide
        self._actuation_point = actuationPoint
        self._closure = WeArtCommon.defaultClosure
        self._abduction = WeArtCommon.defaultAbduction
    
    def OnMessageReceived(self, message: WeArtMessage):
        if message.getID() == TrackingMessage.ID:
            # it means that message is actually a TrackingMessage
            self._closure = message.GetClosure(self._handSide, self._actuation_point)
            self._abduction = message.GetAbduction(self._handSide, self._actuation_point)

    def GetClosure(self):
        return self._closure
    
    def GetAbduction(self):
        return self._abduction

__all__ = ['WeArtThimbleTrackingObject']
